
# This file was generated by 'versioneer.py' (0.18) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "branch": "0.11.0)",
 "date": "2020-05-10T18:17:21+0100",
 "dirty": false,
 "error": null,
 "full-revisionid": "3028710b5652aa98dd3409ee669b7e6e428d2525",
 "repository": "https://github.com/metawards/MetaWards",
 "version": "0.11.0"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)




from typing import Union as _Union
from typing import List as _List
from ..utils._get_functions import MetaFunction, accepts_stage

__all__ = ["extract_custom",
           "build_custom_extractor"]


def build_custom_extractor(custom_function: _Union[str, MetaFunction],
                           parent_name="__main__") -> MetaFunction:
    """Build and return a custom extractor from the passed
       function. This will wrap 'extract_custom' around
       the function to double-check that the custom
       function is doing everything correctly

       Parameters
       ----------
       custom_function: MetaFunction or str
         This can either be a function, which will be wrapped and
         returned, or it can be a string. If it is a string then
         we will attempt to locate or import the function associated
         with that string. The search order is;

         1. Is this 'metawards.extractors.custom_function'?
         2. Is this 'custom_function' that is already imported'?
         3. Is this a file name in the current path, if yes then
            find the function in that file (either the first function
            called 'extractXXX' or the specified function if
            custom_function is in the form module::function)

        parent_name: str
          This should be the __name__ of the calling function, e.g.
          call this function as build_custom_extractor(func, __name__)

        Returns
        -------
        extractor: MetaFunction
          The wrapped extractor that is suitable for using in the extract
          function.
    """
    print(custom_function)

    if isinstance(custom_function, str):
        print(f"Importing a custom extractor from {custom_function}")

        # we need to find the function
        import metawards.extractors

        # is it metawards.extractors.{custom_function}
        try:
            func = getattr(metawards.extractors, custom_function)
            return build_custom_extractor(func)
        except Exception:
            pass

        # do we have the function in the current namespace?
        import sys
        try:
            func = getattr(sys.modules[__name__], custom_function)
            return build_custom_extractor(func)
        except Exception:
            pass

        # how about the __name__ namespace of the caller
        try:
            func = getattr(sys.modules[parent_name], custom_function)
            return build_custom_extractor(func)
        except Exception:
            pass

        # how about the __main__ namespace (e.g. if this was loaded
        # in a script)
        try:
            func = getattr(sys.modules["__main__"], custom_function)
            return build_custom_extractor(func)
        except Exception:
            pass

        # can we import this function as a file - need to check that
        # the user hasn't written this as module::function
        if custom_function.find("::") != -1:
            parts = custom_function.split("::")
            func_name = parts[-1]
            func_module = "::".join(parts[0:-1])
        else:
            func_name = None
            func_module = custom_function

        from ..utils._import_module import import_module

        module = import_module(func_module)

        if module is None:
            # we cannot find the extractor
            print(f"Cannot find the extractor '{custom_function}'."
                  f"Please make sure this is spelled correctly and "
                  f"any python modules/files needed are in the "
                  f"PYTHONPATH or current directory")
            raise ImportError(f"Could not import the extractor "
                              f"'{custom_function}'")

        if func_name is None:
            # find the first function that starts with 'extract'
            import inspect
            for name, value in inspect.getmembers(module):
                if name.startswith("extract"):
                    if hasattr(value, "__call__"):
                        # this is a function
                        return build_custom_extractor(getattr(module, name))

            print(f"Could not find any function in the module "
                  f"{custom_function} that has a name that starts "
                  f"with 'extract'. Please manually specify the "
                  f"name using the '{custom_function}::your_function syntax")

            raise ImportError(f"Could not import the extractor "
                              f"{custom_function}")

        else:
            if hasattr(module, func_name):
                return build_custom_extractor(getattr(module, func_name))

            print(f"Could not find the function {func_name} in the "
                  f"module {func_module}. Check that the spelling "
                  f"is correct and that the right version of the module "
                  f"is being loaded.")
            raise ImportError(f"Could not import the extractor "
                              f"{custom_function}")

    if not hasattr(custom_function, "__call__"):
        print(f"Cannot build an extractor for {custom_function} "
              f"as it is missing a __call__ function, i.e. it is "
              f"not a function.")
        raise ValueError(f"You can only build custom extractors for "
                         f"actual functions... {custom_function}")

    print(f"Building a custom extractor for {custom_function}")

    return lambda **kwargs: extract_custom(custom_function=custom_function,
                                           **kwargs)


def extract_custom(custom_function: MetaFunction,
                   stage: str, **kwargs) -> _List[MetaFunction]:
    """This returns the default list of 'output_XXX' functions that
       are called in sequence for each iteration of the model run.
       This provides a custom extractor that uses
       'custom_function' passed from the user. This makes
       sure that if 'stage' is not handled by the custom function,
       then the "extract_default" functions for that stage
       are correctly called for all stages except "analyse"

       Parameters
       ----------
       custom_function: MetaFunction
         A custom user-supplied function that returns the
         functions that the user would like to be called for
         each step.
       stage: str
         The stage of the day/model

       Returns
       -------
       funcs: List[MetaFunction]
         The list of functions that will be called in sequence
    """

    kwargs["stage"] = stage

    if custom_function is None:
        from ._extract_default import extract_default
        return extract_default(**kwargs)

    elif stage == "analyse" or accepts_stage(custom_function):
        # most custom functions operate at the 'analyse' stage,
        # so extractors that don't specify a stage are assumed to
        # only operate here (every other stage is 'extract_default')
        return custom_function(**kwargs)

    else:
        from ._extract_default import extract_default
        return extract_default(**kwargs)



"""
.. currentmodule:: metawards

Classes
=======

.. autosummary::
    :toctree: generated/

    Demographic
    Demographics
    Disease
    Infections
    InputFiles
    Link
    Links
    Network
    Networks
    Node
    Nodes
    OutputFiles
    Parameters
    Population
    Populations
    VariableSet
    VariableSets
    WardInfo
    WardInfos
    Workspace

Functions
=========

.. autosummary::
    :toctree: generated/

    get_version_string

"""

from . import analysis
from . import app
from . import movers
from . import mixers
from . import extractors
from . import iterators
from . import utils
from ._version import get_versions
from ._links import *
from ._nodes import *
from ._wardinfo import *
from ._workspace import *
from ._infections import *
from ._variableset import *
from ._outputfiles import *
from ._networks import *
from ._network import *
from ._population import *
from ._link import *
from ._node import *
from ._inputfiles import *
from ._disease import *
from ._demographics import *
from ._demographic import *
from ._parameters import *
import sys as _sys

if _sys.version_info < (3, 7):
    print("MetaWards requires Python version 3.7 or above.")
    print("Your python is version")
    print(_sys.version)
    _sys.exit(-1)

__all__ = ["get_version_string"]


# import the pyx cython-compiled modules

__manual_version__ = "0.9.0"

_v = get_versions()
__version__ = _v['version']

if __version__.find("untagged") != -1:
    __version__ = __manual_version__

__branch__ = _v['branch']
__repository__ = _v['repository']
__revisionid__ = _v['full-revisionid']
del _v
del get_versions


def get_version_string():
    """Return a version string for metawards which can be printed
       into a file or written out to the screen
    """
    from ._version import get_versions
    v = get_versions()

    if v['version'].find("untagged") != -1:
        # the version couldn't be found - this is likely because this
        # is run as part of github actions or in a disconnected repo.
        # We need to drop back to '__manual_version__'
        v['version'] = __manual_version__

    header = f"metawards version {v['version']}"
    stars = "*" * len(header)

    def center(line, length=48):
        diff = length - len(line)
        if diff <= 0:
            return line
        else:
            left = int(diff/2) * " "
            return left + line

    newline = "\n"

    stars = center(stars)

    lines = []
    lines.append("")
    lines.append(stars)
    lines.append(center(header))
    lines.append(stars)
    lines.append("")

    lines.append(center(f"-- Source information --"))
    lines.append(center(f"repository: {v['repository']}"))
    lines.append(center(f"branch: {v['branch']}"))
    lines.append(center(f"revision: {v['full-revisionid']}"))
    lines.append(center(f"last modified: {v['date']}"))

    if v["dirty"]:
        lines.append("")
        lines.append("WARNING: This version has not been committed to git,")
        lines.append("WARNING: so you may not be able to recover the original")
        lines.append("WARNING: source code that was used to generate "
                     "this run!")

    lines.append("")
    lines.append(center(f"-- Additional information --"))
    lines.append(center(f"Visit https://metawards.org for more information"))
    lines.append(center(f"about metawards, its authors and its license"))
    lines.append("")

    return newline.join(lines)


_system_input = input


def input(prompt: str, default="y"):
    """Wrapper for 'input' that returns 'default' if it detected
       that this is being run from within a batch job or other
       service that doesn't have access to a tty
    """
    import sys

    try:
        if sys.stdin.isatty():
            return _system_input(prompt)
        else:
            print(f"Not connected to a console, so having to use the "
                  f"default ({default})")
            return default
    except Exception as e:
        print(f"Unable to get the input: {e.__class__} {e}")
        print(f"Using the default ({default}) instead")
        return default
